/*
** $Id: conf.c,v 1.2 2002/09/24 22:05:13 pief Exp $
**
** MetalServe common stuff
** Copyright (c) 2002 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Configuration parser functions
**
** $Log: conf.c,v $
** Revision 1.2  2002/09/24 22:05:13  pief
** For some reason I checked in ancient versions, so here are the current
** ones
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "base.h"
#include "strings.h"
#include "conf.h"
#include "conf-internal.h"

/*
** cfgerr(cfgfile, cfgerrno)
**
** Sets a config error code and generates a suitable error string.
*/
void cfgerr(CFGFILE *cfgfile, int cfgerrno)
{
	char errstrbuf[500];

	cfgfile->cfgerrno = cfgerrno;

	switch (cfgerrno)
	{
		case CFG_ERR_OUTOFMEM:
		{
			cfgfile->cfgerrstr = "Out of memory";
			return;
		}
		case CFG_ERR_INVDIRECTIVE:
		{
			sprintf(errstrbuf, "Unknown or invalid directive \"%s\"", cfgfile->dstr);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_NOARG:
		{
			sprintf(errstrbuf, "No argument to directive \"%s\" specified", cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGCOUNT:
		{
			sprintf(errstrbuf, "Invalid number of arguments to directive \"%s\"", cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGTOOSHORT:
		{
			sprintf(errstrbuf, "Argument to directive \"%s\" is too short (%d < %d characters)",
			        cfgfile->d->name, strlen(cfgfile->argstr), cfgfile->d->min);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGTOOLONG:
		{
			sprintf(errstrbuf, "Argument to directive \"%s\" is too long (%d > %d characters)",
			        cfgfile->d->name, strlen(cfgfile->argstr), cfgfile->d->max);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_MISSINGTERMQUOTES:
		{
			sprintf(errstrbuf, "Missing terminating double quotes in argument to directive \"%s\"",
			        cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_INVARG:
		{
			sprintf(errstrbuf, "\"%s\" is no valid argument to directive \"%s\"",
			        cfgfile->argstr, cfgfile->d->name);
			cfgfile->cfgerrstr = errstrbuf;
			return;
		}
		case CFG_ERR_ARGTOOSMALL:
		{
			sprintf(errstrbuf, "Numeric argument to directive \"%s\" is too small (%d < %d)",
			        cfgfile->d->name, *((int *)cfgfile->d->target), cfgfile->d->min);
			cfgfile->cfgerrstr = errstrbuf;
			return;			
		}
		case CFG_ERR_ARGTOOLARGE:
		{
			sprintf(errstrbuf, "Numeric argument to directive \"%s\" is too large (%d > %d)",
			        cfgfile->d->name, *((int *)cfgfile->d->target), cfgfile->d->max);
			cfgfile->cfgerrstr = errstrbuf;
			return;			
		}
		case CFG_ERR_TOOMANYARGS:
		{
			sprintf(errstrbuf, "Too many arguments in argument list to directive \"%s\" (max. allowed: %d)",
			        cfgfile->d->name, cfgfile->d->max);
			cfgfile->cfgerrstr = errstrbuf;
			return;			
		}
		case CFG_ERR_NONE:
		default:
		{
			cfgfile->cfgerrstr = NULL;
		}
	}
}

/*
** cfgfile = cfgopen(filename, flags)
**
** Opens the specified configuration file for reading and returns a CFGFILE
** pointer for use with the other API functions.
**
** Returns TRUE on success or FALSE on failure, in which case the "errno"
** variable will contain the appropriate error code.
*/
CFGFILE *cfgopen(const char *filename, const int flags)
{
	CFGFILE *cfgfile;

	/* Allocate memory for CFGFILE structure */
	cfgfile = malloc(sizeof(CFGFILE));
	if (!cfgfile)
	{
		errno = ENOMEM;
		return NULL;
	}

	memset(cfgfile, 0, sizeof(CFGFILE));

	/* Open the specified configuration file */
	cfgfile->fp = fopen(filename, "r");
	if (!cfgfile->fp)
	{
		free(cfgfile);
		return NULL;
	}

	/* Set flags and initial configuration error code */
	cfgfile->flags    = flags;
	cfgfile->cfgerrno = CFG_ERR_NONE;

	return cfgfile;
}

/*
** cfgclose(cfgfile)
**
** Closes the specified configuration file and releases any resources used.
**
** After use, the variable cfgfile should be set to NULL, if chances are that
** it could be reused at a later time.
*/
void cfgclose(CFGFILE *cfgfile)
{
	if (!cfgfile) return;
	
	if (cfgfile->fp) fclose(cfgfile->fp);

	free(cfgfile);
}

/*
** rc = cfgread(cfgfile, directives)
**
** Processes a configuration file using the supplied array of CFGDIRECTIVE
** structures.
**
** Returns TRUE on success or FALSE on failure, in which case _cfgerrno
** contains an error code for which a message may be obtained using
** cfgerror().
*/
BOOL cfgread(CFGFILE *cfgfile, CFGDIRECTIVE *directives)
{
	if (!directives)
	{
		cfgerr(cfgfile, CFG_ERR_INVDIRECTIVE);
		return FALSE;
	}

	/* Process line for line */
	while (fgets(cfgfile->linebuf, sizeof(cfgfile->linebuf), cfgfile->fp))
	{
		cfgfile->lineno++;

		/* Strip whitespace, CRs and LFs */
		cfgfile->linep = trim_line(cfgfile->linebuf);

		/* Skip empty lines and comment lines */
		if ( (*cfgfile->linep == '#') || (*cfgfile->linep == '\0') )
			continue;

		/* Isolate directive */
		cfgfile->dstr = cfgfile->linep;
		cfgfile->argstr = cfgfile->linep + strcspn(cfgfile->linep, " \t\0");

		/* Skip whitespace to argument */
		if (*cfgfile->argstr != '\0')
		{
			*cfgfile->argstr = '\0';
			cfgfile->argstr++;
			cfgfile->argstr += strspn(cfgfile->argstr, " \t");
		}
		else cfgfile->argstr = NULL;

//		printf("line %d directive: [%s] arg: [%s]\n", cfgfile->lineno, cfgfile->dstr, cfgfile->argstr);
		
		/* Look up directive */
		for (cfgfile->d = directives; cfgfile->d->name; cfgfile->d++)
		{
			if (!strcasecmp(cfgfile->dstr, cfgfile->d->name))
				break;
		}

		/* Did we get a match? */
		if (!cfgfile->d->name)
		{
			/* If the flag CFG_FLAGS_IGNORE_UNKNOWN was specified to cfgopen(),
			   ignore the unknown directive */
			if (cfgfile->flags & CFG_FLAGS_IGNORE_UNKNOWN) continue;

			/* Otherwise report an error message */
			cfgerr(cfgfile, CFG_ERR_INVDIRECTIVE);
			return FALSE;	
		}

		/* Check whether no argument was specified */
		if (!cfgfile->argstr)
		{
			cfgerr(cfgfile, CFG_ERR_NOARG);
			return FALSE;	
		}

		/* If no argument parser function was defined, skip this directive */
		if (!cfgfile->d->apfunc)
			continue;

		/* Otherwise call that callback function */
		if (!cfgfile->d->apfunc(cfgfile, cfgfile->d, NULL, cfgfile->argstr))
			return FALSE;
	}
	if (errno) return FALSE;

	return TRUE;
}

/*
** cfgerr = cfgerrno(CFGFILE *cfgfile)
**
** Returns the error code of the last configuration error that occured while
** reading the specified configuration file, which may be CFG_ERR_NONE.
*/
int cfgerrno(CFGFILE *cfgfile)
{
	return cfgfile->cfgerrno;
}

/*
** errstr = cfgerrstr(CFGFILE *cfgfile)
**
** Returns a string describing the last configuration error that occured while
** reading the specified configuration file, which may be a NULL pointer if
** no error occured.
*/
char *cfgerrstr(CFGFILE *cfgfile)
{
	return cfgfile->cfgerrstr;
}

/*
** lineno = cfglineno(CFGFILE *cfgfile)
**
** Returns the number of the last line that was read before a configuration
** error occured.
*/
int cfglineno(CFGFILE *cfgfile)
{
	return cfgfile->lineno;
}

/*
** Argument parser function for C-style strings
*/
APFUNC(argparser_string)
{
	/* Check for empty string */
	if ( (*arg == '"') && (*(arg+1) == '"') )
	{
		/* Are empty strings allowed? */
		if (current->min == 0)
		{
			*((char **)current->target) = NULL;
			return TRUE;
		}
		else
		{
			cfgerr(cfgfile, CFG_ERR_NOARG);
			return FALSE;
		}
	}

	/* Check argument length */
	if (strlen(arg) < current->min)
	{
		cfgerr(cfgfile, CFG_ERR_ARGTOOSHORT);
		return FALSE;
	}
	if (strlen(arg) > current->max)
	{
		cfgerr(cfgfile, CFG_ERR_ARGTOOLONG);
		return FALSE;
	}

	/* Assign copy of (unescaped) string argument */
	if (current->flags == APFUNC_FLAG_DONT_UNESCAPE)
		*((char **)current->target) = strdup(arg);
	else
		*((char **)current->target) = unescape(arg);

	return TRUE;
}

/*
** Argument parser function for switches. The keywords "yes", "on" and "true"
** cause the configuration variable to be set to 1, the keywords "no", "off"
** and "false" set it to 0. Other arguments cause the parser to fail.
*/
APFUNC(argparser_switch)
{
	/* Check for recognized keywords */
	if ( (strcmp(arg, "yes") == 0) || (strcmp(arg, "on") == 0) || (strcmp(arg, "true") == 0) )
	{
		*((BOOL *)current->target) = TRUE;
		return TRUE;
	}
	else if ( (strcmp(arg, "no") == 0) || (strcmp(arg, "off") == 0) || (strcmp(arg, "false") == 0) )
	{
		*((BOOL *)current->target) = FALSE;
		return TRUE;
	}

	/* An unrecognized keyword was specified */
	cfgerr(cfgfile, CFG_ERR_INVARG);
	return FALSE;
}

/*
** Argument parser function for directives that take a command and optional arguments.
** Handles quoted and escaped command names properly and stores the arguments part
** in once (unescaped if the APFUNC_FLAG_DONT_UNESCAPE flag has been set). Handy for
** a safe fork() / execve() combination instead of a system() call.
*/
APFUNC(argparser_command)
{
	BOOL escaped, quoted;
	char *p;

	/* Check for empty string */
	if ( (*arg == '"') && (*(arg+1) == '"') )
	{
		cfgerr(cfgfile, CFG_ERR_NOARG);
		return FALSE;
	}

	/* Extract command */
	escaped = FALSE;
	quoted = FALSE;

	for (p = arg; *p; p++)
	{
		/* "Space" seperates the program name from the arguments. It may be
		    quoted or escaped if it needs to be part of the program name. */
		if (*p == ' ')
		{
			if (escaped) escaped = FALSE;
			else if (!quoted) break;
		}

		/* Toggle quote mode? */
		if (*p == '"') quoted = 1 - quoted;

		/* Enable escape mode? */
		if (*p == '\\') escaped = TRUE;
	}

	/* If we're still in quoted mode at this place, we've reached the end of the
	   argument and we miss terminating double quotes. */
	if (quoted)
	{
		cfgerr(cfgfile, CFG_ERR_MISSINGTERMQUOTES);
		return FALSE;
	}

	*p++ = '\0';
	((CFGCOMMAND *)current->target)->program = strdup(arg);

	/* Assign copy of (unescaped) string argument */
	if (current->flags == APFUNC_FLAG_DONT_UNESCAPE)
		((CFGCOMMAND *)current->target)->args = strdup(p);
	else
		((CFGCOMMAND *)current->target)->args = unescape(p);

	return TRUE;
}

/*
** Argument parser function for integer values
*/
APFUNC(argparser_int)
{
	char *p;
	int value;

	/* Attempt to convert string representation to a numeric value */
	value = strtol(arg, &p, 10);
	if (*p != '\0')
	{
		cfgerr(cfgfile, CFG_ERR_INVARG);
		return FALSE;
	}

	/* Store converted value already here so that the following checks
	   have the value available in their error messages */
	*((int *)current->target) = value;

	/* Check whether converted value is within the allowed range */
	if (value < current->min)
	{
		cfgerr(cfgfile, CFG_ERR_ARGTOOSMALL);
		return FALSE;
	}
	if (value > current->max)
	{
		cfgerr(cfgfile, CFG_ERR_ARGTOOLARGE);
		return FALSE;
	}

	return TRUE;
}

/*
** Argument parser for string lists
*/
APFUNC(argparser_stringlist)
{
        char **ptr_array;
        int ptrs_left;

	char *nextarg;
	int quoted, escaped;

	/* Allocate pointer array for variable-argument list */
	ptr_array = calloc(MAX_LIST_ENTRIES + 1, sizeof(char *));
	if (!ptr_array)
	{
		cfgerr(cfgfile, CFG_ERR_OUTOFMEM);
		return FALSE;
	}
	ptrs_left = MAX_LIST_ENTRIES;
	*((char ***)current->target) = ptr_array;

	/* Process list string by string */
	for (nextarg = arg; arg; arg = nextarg)
	{
		/* Determine next string in list */
		quoted = FALSE;
		escaped = FALSE;
		while (*nextarg != '\0')
		{
			/* We use the "space" character to determine the next
			   argument. "Space" may be either quoted or escaped
			   if it needs to be included into an argument. */
			if (*nextarg == ' ')
			{
				if (escaped) escaped = FALSE;
				else if (!quoted) break;
			}
			/* "Tab" is recognized as argument seperator as well,
			   however unlike "Space" it can be quoted, but not
			   escaped. */
			if ( (*nextarg == '\t') && (!quoted) ) break;

			/* Toggle quote mode? */
			if (*nextarg == '"') quoted = 1 - quoted;

			/* Enable escape mode? */
			if (*nextarg == '\\') escaped = TRUE;

			nextarg++;
		}

		if (*nextarg != '\0')
		{
			*nextarg = 0;
			nextarg++;
		}
		else nextarg = NULL;

		/* Store pointer to an unescaped copy of this argument in the pointer array */
		*ptr_array = unescape(arg);
		ptrs_left--;
		ptr_array++;

		/* If all pointers have been used, bail out and fail */
		if ( (!ptrs_left) || ((MAX_LIST_ENTRIES - ptrs_left) > current->max) )
		{
			cfgerr(cfgfile, CFG_ERR_TOOMANYARGS);
			return FALSE;
		}
	}

	/* Store an additional NULL pointer so that calling functions know where the
	   pointer array ends */
	*ptr_array = NULL;

	return TRUE;
}
