/*
** $Id: conf.h,v 1.2 2002/09/24 22:05:13 pief Exp $
**
** MetalServe common stuff
** Copyright (c) 2002 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Header file for configuration parser functions
**
** $Log: conf.h,v $
** Revision 1.2  2002/09/24 22:05:13  pief
** For some reason I checked in ancient versions, so here are the current
** ones
**
*/

#ifndef _CONF_H
#define _CONF_H

/*
** The CFGFILE structure is for pconf-internal use only, but is declared here because
** we need it in the prototypes.
*/
typedef struct _cfgfile CFGFILE;

/*
** Forward declaration of CFGDIRECTIVE structure
*/
typedef struct _cfgdirective CFGDIRECTIVE;

/*
** Argument parser functions receive the following arguments, which are hidden by the
** macro APFUNC_ARGS:
**
** - "cfgfile": the currently processed cfgfile (used for cfgerr() only)
** - "current": the currently processed directive (which includes the target variable)
** - "parent" : the parent directive, if available
** - "arg"    : the remainder of the config line read (ie. the argument(s) to the directive)
*/
#define APFUNC_ARGS CFGFILE *cfgfile, const CFGDIRECTIVE *current, CFGDIRECTIVE *parent, char *arg

/*
** They return an "int" value, which is TRUE on success or FALSE if the parser should abort.
** Positive return codes are reserved for internal use.
**
** These semantics are hidden behind the macro APFUNC which should be used for the
** declaration and definition of argument parser functions.
*/
#define APFUNC(_name) int _name(APFUNC_ARGS)

/*
** These are the prototypes for the built-in argument parser functions.
*/
APFUNC(argparser_string);
APFUNC(argparser_switch);
APFUNC(argparser_command);
APFUNC(argparser_int);
APFUNC(argparser_stringlist);

/*
** A CFGDIRECTIVE structure specifies a "known" configuration directive.
*/
struct _cfgdirective
{
	char	*name;					/* Name of directive */
	int	(*apfunc)(APFUNC_ARGS);			/* Function responsible for argument parsing */
	void	*target;				/* Target variable to be modified by apfunc */
	int	min,					/* Minimum value (may be used by apfunc) */
		max;					/* Maximum value (may be used by apfunc) */
	int	flags;					/* Flags as defined below (may be used by apfunc) */
};

/*
** Definitions for the "flags" field above
*/
#define APFUNC_FLAG_NONE		0		/* No flags */
#define APFUNC_FLAG_DONT_UNESCAPE	1		/* Strings should not be unescaped */

/*
** An array of CFGDIRECTIVE structures defines the entire configuration file. It must be terminated
** with the LAST_CFGDIRECTIVE macro.
*/
#define LAST_CFGDIRECTIVE { NULL, NULL, NULL, 0, 0 }

/*
** A CFGCOMMAND structure contains the parsed command name and the arguments part, handy for use
** with a safe fork() / execve() combination instead of a system() call.
*/
typedef struct _cfgcommand
{
	char	*program;				/* The parsed program name (always unescaped) */
	char	*args;					/* The arguments part (by default also unescaped) */
} CFGCOMMAND;

/*
** Flags to the cfgopen() function
*/
#define CFG_FLAGS_NONE			0		/* No special flags set */
#define CFG_FLAGS_IGNORE_UNKNOWN	1		/* Ignore unknown directives (instead of reporting errors) */

/*
** Maximum number of entries in variable-argument lists such as string lists
*/
#define MAX_LIST_ENTRIES 30

/*
** Configuration error code definitions (for use in own argument parser functions)
*/
enum _cfgerrcodes
{
        CFG_ERR_NONE,                                   /* No error */
        CFG_ERR_OUTOFMEM,                               /* Out of memory */
        CFG_ERR_INVDIRECTIVE,                           /* Invalid/unknown directive */
        CFG_ERR_NOARG,                                  /* No argument specified */
        CFG_ERR_ARGCOUNT,                               /* Invalid number of arguments */
        CFG_ERR_ARGTOOSHORT,                            /* Argument too short */
        CFG_ERR_ARGTOOLONG,                             /* Argument too long */
	CFG_ERR_MISSINGTERMQUOTES,			/* Missing terminating double quotes */
        CFG_ERR_INVARG,                                 /* Invalid argument */
        CFG_ERR_ARGTOOSMALL,                            /* Numeric argument too small */
        CFG_ERR_ARGTOOLARGE,                            /* Numeric argument too large */
        CFG_ERR_TOOMANYARGS                             /* Too many arguments in a string list */
};

/*
** Main API prototypes.
*/
CFGFILE *cfgopen(const char *filename, const int flags);
void cfgclose(CFGFILE *cfgfile);
BOOL cfgread(CFGFILE *cfgfile, CFGDIRECTIVE *directives);
int cfgerrno(CFGFILE *cfgfile);
char *cfgerrstr(CFGFILE *cfgfile);
int cfglineno(CFGFILE *cfgfile);

/*
** For own argument parser functions only
*/
void cfgerr(CFGFILE *cfgfile, int cfgerrno);

#endif
