/*
** $Id: makelist.c,v 1.6 2002/10/19 20:51:38 pief Exp $
**
** makelist - MetalServe file list generator
** Copyright (c) 2002 by Pieter Hollants <contact@metalserve.org>
**
** This program is licensed under the GNU General Public License, version 2,
** as published by the Free Software Foundation and available in the file
** COPYING and the Internet location http://www.gnu.org/licenses/gpl.html.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
** or FITNESS FOR A PARTICULAR PURPOSE.
**
** Main program
**
** $Log: makelist.c,v $
** Revision 1.6  2002/10/19 20:51:38  pief
** Added missing #include's to get makelist compiling under FreeBSD
** Added support for the PACKAGE_NAME and PACKAGE_VERSION defines set by the
** top-level configure script
**
** Revision 1.5  2002/10/01 07:54:54  pief
** Moved rotation delay into #define in makelist.h
**
** Revision 1.4  2002/10/01 07:52:46  pief
** Fixed bug in insert_scannode() that caused directories to be unsorted
** when inserted in a certain order that never happened in reality (mentioned
** in metalserve.org news on 20092002, forgot to reactivate the fix)
**
** Revision 1.3  2002/09/30 22:19:40  pief
** Fixes due to wrong CVS commit
**
** Revision 1.2  2002/09/30 22:03:47  pief
** Implemented a progress display with a rotating cursor, which is displayed
** only if a tty is attached, and if so, also respects the terminal's width.
** Made metalserve more verbose in general.
** Added -q command line switch to make metalserve shut up again =)
** Moved code to convert a pure seconds value to minutes and seconds to a
** new common/misc.c routine named get_mins_and_secs()
**
** Revision 1.1  2002/09/24 21:46:03  pief
** Added first version of makelist main files
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <libgen.h>

#include "../lib/getopt.h"

#include "../common/base.h"
#include "../common/list.h"
#include "../common/conf.h"
#include "../common/misc.h"

#include "makelist.h"
#include "mpegaudio.h"
#include "waveaudio.h"

const char *_prgbanner =
	"%s (%s v%s) - file list generator\n"
	"Copyright (c) 2002 by Pieter Hollants <contact@metalserve.org> - http://www.metalserve.org\n\n";

char *_prgname;

BOOL _tty_attached = FALSE;
int _terminal_width;

const char *_rotation_chars = "|/-\\";
#define ROTATION_CHARS_LEN 4
int _rotation = 0;
static struct timeval _last_rotation, _now_tv;

/*
** Command line options
*/
static struct option long_options[] =
{
	{ "configfile",		required_argument,	0, 'f' },
	{ "help",		no_argument,		0, 'h' },
	{ "quiet",		no_argument,		0, 'q' },
	{ "serverroot",		required_argument,	0, 's' },
	{ "version",		no_argument,		0, 'V' },
	{ 0, 0, 0, 0 }
};
const char *_cmdlineopts = "fhqsV";

const char *_cmdlinehelp =
	"This program is licensed under the GNU General Public License, version 2.\n"
	"See the file COPYING or visit http://www.gnu.org/licenses/gpl.html for details.\n\n"

	"Usage: %s [options]\n\n"

	"Options:\n"
	"  -h, --help                      display this information\n"
	"  -f FILE, --configfile=FILE      specify alternative configuration file\n"
	"  -s PATH, --serverroot=PATH      override configured serverroot (root directory to descend)\n"
	"  -q, --quiet                     suppress output (for use when started by a cron job)\n"
	"  -V, --version                   print version and exit\n";

BOOL _quiet = FALSE;

/*
** Configuration file variables and directives
*/
char *_cfgfile_name;

char *_logfile_name;
char *_serverroot;
char *_compress_command;
BOOL *_keep_uncompressed_list;
char **_include_extensions;
char *_list_header;
BOOL *_dir_overview;
char **_nickname;

CFGDIRECTIVE directives[] =
{
	/* directive name, argparser function, target variable, arg restrictions, required directive? */
	{ "logfile",			argparser_string,	&_logfile_name,			0, FILENAME_MAX,	APFUNC_FLAG_NONE },
	{ "serverroot",			argparser_string,	&_serverroot,			0, PATH_MAX,		APFUNC_FLAG_NONE },
	{ "compress_command",		argparser_string,	&_compress_command,		0, FILENAME_MAX,	APFUNC_FLAG_DONT_UNESCAPE },
	{ "keep_uncompressed_list",	argparser_switch,	&_keep_uncompressed_list,	0, 0,			APFUNC_FLAG_NONE },
	{ "include_extensions",		argparser_stringlist,	&_include_extensions,		0, MAX_LIST_ENTRIES,	APFUNC_FLAG_NONE },
	{ "list_header",		argparser_string,	&_list_header,			0, FILENAME_MAX,	APFUNC_FLAG_NONE },
	{ "dir_overview",		argparser_switch,	&_dir_overview,			0, 0,			APFUNC_FLAG_NONE },
	{ "nickname",			argparser_stringlist,	&_nickname,			0, MAX_LIST_ENTRIES,	APFUNC_FLAG_NONE },
	LAST_CFGDIRECTIVE
};

/*
** Filelist variables
*/
LIST *_filelist;
time_t _starttime, _now;
int _mins_passed, _secs_passed;
int _totalkbytes, _totalfiles;

char _filename[NAME_MAX];
FILE *_fd = NULL;

unsigned char _buf[BUF_SIZE];
int _pad;

/*
** rc = parse_cmdline_args(argc, argv)
**
** Parses the program's command line arguments.
**
** Returns TRUE if the program should continue to run or FALSE if the
** program should quit.
*/
static BOOL parse_cmdline_args(int argc, char **argv)
{
	int c, option_index = 0;

	while (TRUE)
	{
		c = getopt_long(argc, argv, _cmdlineopts, long_options, &option_index);
		if (c == -1) break;

		switch (c)
		{
			case 'f':
			{
				_cfgfile_name = strdup(optarg);
				break;
			}
			case 'h':
			{
				printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION);
				printf(_cmdlinehelp, _prgname);
				return FALSE;
			}
			case 'q':
			{
				_quiet = TRUE;
				break;	
			}
			case 's':
			{
				_serverroot = strdup(optarg);
				break;
			}
			case 'V':
			{
				printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION);
				printf("Compiled %s %s\n", __DATE__, __TIME__);
				return FALSE;
			}
			case '?':
			{
				/* getopt_long already printed an error message */
				return FALSE;
			}
		}
	}

	return TRUE;
}

/*
** rc = parse_cfgfile()
**
** Parses the MetalServe configuration file.
**
** Returns TRUE if the program should continue to run or FALSE if the
** program should quit.
*/
static BOOL parse_cfgfile(void)
{
	CFGFILE *cfgfile = NULL;
	char *errstr;
	char *old_serverroot;
	int attempt;

	/* Attempt to open configuration file at one of two locations */
	for (attempt = 0; attempt < 2; attempt++)
	{
		/* Try the system-wide location first */
		if (attempt == 0)
			_cfgfile_name = DEFAULT_CFGFILE_NAME;

		/* Then try the home dir of the user we're running under */
		else
		{
			char *homedir;

			/* Get path of current user's home directory and allocate memory for composed filename */
			homedir = getenv("HOME");
			_cfgfile_name = malloc(PATH_MAX);
			if ( (!homedir) || (!_cfgfile_name) )
			{
				fprintf(stderr, "%s: could not compose full filename of \"%s\" configuration file!\n",
					_prgname, HOMEDIR_CFGFILE_NAME);
				return FALSE;
			}
			
			snprintf(_cfgfile_name, PATH_MAX, "%s/%s", homedir, HOMEDIR_CFGFILE_NAME);
		}

		/* Now try to actually open the configuration file */
		errno = 0;
		cfgfile = cfgopen(_cfgfile_name, CFG_FLAGS_IGNORE_UNKNOWN);
		if ( (cfgfile) || ( (!cfgfile) && (errno != ENOENT) ) )
			break;
	}
	
	/* Couldn't we open it for some reason? */
	if (!cfgfile)
	{
		/* File did not exist at both locations */
		if (errno == ENOENT)
		{
			fprintf(stderr, "%s: missing configuration file at either \"%s\" or \"%s\"!\n",
				_prgname, DEFAULT_CFGFILE_NAME, HOMEDIR_CFGFILE_NAME);
		}
		/* Some other serious error */
		else
		{	
			fprintf(stderr, "%s: error opening configuration file \"%s\":\n%s!\n",
				_prgname, _cfgfile_name, strerror(errno));
		}

		return FALSE;
	}

	/* Process configuration directives (but remember _serverroot eventually specified on command line) */
	old_serverroot = _serverroot;
	if (!cfgread(cfgfile, directives))
	{
		if (cfgerrno(cfgfile)) errstr = cfgerrstr(cfgfile);
		else errstr = strerror(errno);
		fprintf(stderr, "%s: error while reading line %d of \"%s\":\n%s!\n",
			_prgname, cfglineno(cfgfile), _cfgfile_name, errstr);
		return FALSE;
	}
	if (old_serverroot) _serverroot = old_serverroot;

	cfgclose(cfgfile);

	return TRUE;
}

/*
** width = determine_terminal_width()
**
** Determine the width of the attached terminal.
*/
static int determine_terminal_width(void)
{
	struct winsize wsz;

	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &wsz) < 0)
	{
		fprintf(stderr, "%s: could not determine terminal width:\n%s!\n",
			_prgname, strerror(errno));
		return 0;
	}

	return wsz.ws_col;
}

/*
** rc = start_filelist()
**
** Composes the name for the filelist, opens it and writes the MetalServe header.
*/
static BOOL start_filelist(void)
{
	/* Generate file list name */
	strftime(_buf, sizeof(_buf), "%m-%d-%Y", localtime(&_starttime));
	sprintf(_filename, "%s-[%s].txt", *_nickname, _buf);

	/* Initialize counters */
	_totalkbytes = 0;
	_totalfiles = 0;

	/* Open file list for writing */
	_fd = fopen(_filename, "w");
	if (!_fd)
	{
		fprintf(stderr, "%s: error opening file \"%s\" for writing:\n%s!\n",
			_prgname, _filename, strerror(errno));
		return FALSE;
	}

	/* Write MetalServe header */
	fprintf(_fd, "\n");
	_pad = (83 - strlen(*_nickname)) / 2;
	fprintf(_fd, "%*c-{{ List of %s's shared files }}-\n", _pad, ' ', *_nickname);
	fprintf(_fd, "\n");

	strftime(_buf, sizeof(_buf), "%a %d-%B-%Y %T GMT%z", localtime(&_starttime));
	_pad = (83 - strlen(_buf) - strlen(PACKAGE_VERSION)) / 2;
	fprintf(_fd, "%*cLast updated on %s by MetalServe %s\n", _pad, ' ', _buf, PACKAGE_VERSION);
	fprintf(_fd, "%*chttp://www.metalserve.org\n", 45, ' ');

	fprintf(_fd, "\n");
	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");

	/* Include auxiliary list header, if configured */
	if (_list_header)
	{
		FILE *headerfd;
		
		/* Attempt to open header */
		headerfd = fopen(_list_header, "r");
		if (headerfd)
		{
			/* Copy header, line by line */
			while (fgets(_buf, sizeof(_buf), headerfd))
				fputs(_buf,_fd);
			if (ferror(headerfd))
			{
				fprintf(stderr, "%s: error while reading from list_header file \"%s\":\n%s!\n",
					_prgname, _list_header, strerror(errno));
				return FALSE;					
			}

			/* Print additional seperator for instructions below */
			fprintf(_fd, "\n");
			fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
			fprintf(_fd, "\n");
		}
		else
		{
			fprintf(stderr, "%s: error opening list_header file \"%s\" for reading:\n%s!\n",
				_prgname, _list_header, strerror(errno));
			return FALSE;		
		}
	}

	/* Write instructions header */
	fprintf(_fd, "MetalServe instructions:\n\n");

	fprintf(_fd, "To receive a file from me, type \"!%s <exact filename>\" in a channel window\n", *_nickname);
	fprintf(_fd, "or copy and paste one of the lines below into a channel I'm in.\n\n");

	fprintf(_fd, "For example: !%s Metal rules.mp3\n\n", *_nickname);

	fprintf(_fd, "Type @%s-help to receive additional help.\n\n", *_nickname);

	return TRUE;
}

/*
** rc = dirent_selector(dirent)
**
** Callback function called by glibc's scandir() function. Examines the passed
** dirent structure and returns zero for entries that should be excluded.
*/
static int dirent_selector(const struct dirent *entry)
{
	/* Exclude "." and ".." entries */
	if (entry->d_name[0] == '.')
	{
		if (entry->d_name[1] == '\0') return 0;
		if ( (entry->d_name[1] == '.') && (entry->d_name[2] == '\0') ) return 0;
	}

	/* We can't put the include_extensions code here since entry->d_type
	   is not guaranteed to be available in a usable manner on all platforms. */

	return 1;
}

/*
** insert_scannode(list, node)
**
** Inserts the specified SCANNODE at the "correct" position in "list". The
** list is sorted so that directories come alphabetically first, followed
** by the files in alphabetical order.
*/
static void insert_scannode(LIST *list, SCANNODE *node)
{
	SCANNODE *current;
	BOOL zero_ok = FALSE;

	for (current = (SCANNODE *)list->first; current; current = (SCANNODE *)current->next)
	{
		/* If the node to be inserted is a directory, stop search as soon
		   as the first file is reached */
		if ( (IS_DIR(node)) && (!(IS_DIR(current))) )
		{
			current = (SCANNODE *)current->prev;
			zero_ok = TRUE;
			break;
		}

		/* Likewise if the node to be inserted is a file, skip the directory
		   nodes at the beginning of the list */
		if ( !(IS_DIR(node)) && (IS_DIR(current)) )
			continue;

		/* Find the alphabetically correct place */
		if (strcasecmp(node->name, current->name)<0)
		{
			current = (SCANNODE *)current->prev;
			zero_ok = TRUE;
			break;
		}
	}

	/* Should we have passed the last node, we need to go back for it for
	   insert_node() to work properly */
	if ((!current) && (!zero_ok)) current = (SCANNODE *)list->last;

	/* Aha, we found a suitable place */
	insert_node(list, (NODE *)current, (NODE *)node);
}

/*
** rc = process_directory(list, dirname)
**
** Processes the specified directory and adds SCANNODEs for matching files and
** directories to "list". Skips those files that don't match the configured
** include extensions.
*/
static BOOL process_directory(LIST *list, char *dirname)
{
	struct dirent **entries;
	int i, num_entries;

	/* Scan directory */
	num_entries = scandir(dirname, &entries, dirent_selector, alphasort);
	if (num_entries < 0)
	{
		fprintf(stderr, "%s: error while scanning directory \"%s\":\n%s!\n",
			_prgname, dirname, strerror(errno));
		return FALSE;
	}

	for (i = 0; i < num_entries; i++)
	{
		char filename[NAME_MAX];
		struct stat statbuf;
		SCANNODE *node;

		/* If a tty is attached, update rotating cursor four times each second */
		if (_tty_attached)
		{
			gettimeofday(&_now_tv, NULL);
			if ((unsigned long)_now_tv.tv_usec - (unsigned long)_last_rotation.tv_usec > ROTATION_DELAY)
			{
				printf("\r%c\r", _rotation_chars[_rotation]);
				fflush(stdout);

				_last_rotation.tv_usec = _now_tv.tv_usec;
				if (_rotation < ROTATION_CHARS_LEN - 1)
					_rotation++;
				else _rotation = 0;
			}
		}

		/* Examine entry type and file size */
		snprintf(filename, NAME_MAX-1, "%s/%s", dirname, entries[i]->d_name);
		if (stat(filename, &statbuf) < 0)
		{
			fprintf(stderr, "%s: error stat()'ing file \"%s\":\n%s!\n",
				_prgname, filename, strerror(errno));
			return FALSE;
		}

		/* Was an "include_extensions" configuration directive given and is this a directory? */
		if ( !(S_ISDIR(statbuf.st_mode)) && (_include_extensions) )
		{
			char *current_extension, **extension;
			BOOL match = FALSE;

			/* If so, determine the current file's extension... */
			current_extension = strrchr(entries[i]->d_name, '.');
			if (!current_extension)
				continue;

			/* ...and compare with the configured file extensions */
			for (extension = _include_extensions; (!match) && (*extension); extension++)
			{
				if (!strcasecmp(current_extension, *extension))
					match = TRUE;
			}
			if (!match) continue;
		}

		/* Allocate SCANNODE structure for this entry */
		node = (SCANNODE *)calloc(sizeof(SCANNODE), 1);
		if (!node)
		{
			fprintf(stderr, "%s: out of memory while processing entry \"%s\"!\n",
				_prgname, filename);
			return FALSE;
		}
		node->name = strdup(entries[i]->d_name);
		node->fullname = strdup(filename);

		if (S_ISDIR(statbuf.st_mode))
		{
			/* If a tty is attached, print progress line for this subdirectory */
			if (_tty_attached)
			{
				printf("\r%c %-*.*s\r", _rotation_chars[_rotation],
				       _terminal_width - 3, _terminal_width - 3, filename);
				fflush(stdout);
			}

			/* Allocate list */
			node->contents = allocate_list();
			if (!node->contents)
			{
				fprintf(stderr, "%s: out of memory while processing directory \"%s\"!\n",
					_prgname, filename);
				return FALSE;
			}

			insert_scannode(list, node);

			/* Recursively call ourself for the contents of the directory */
			if (!process_directory(node->contents, filename))
				return FALSE;
		}
		else
		{
			FILE *file;
			char *extension;

			/* Examine file */
			node->size = statbuf.st_size;

			file = fopen(filename, "r");
			if (!file)
			{
				node->infotype = INFOTYPE_NONE;
				insert_scannode(list, node);
				continue;
			}
			
			/* Filetype detection goes here */
			extension = strrchr(node->name, '.');
			if (!(strncasecmp(extension, ".wav", 4)))
				examine_waveaudio_file(file, filename, node);
			else
				examine_mpegaudio_file(file, filename, node);

			/* Close file and insert node into filelist */
			fclose(file);
			insert_scannode(list, node);
		}
	}

	return TRUE;
}

/*
** write_dir_overview_entry(startlist);
**
** Subfunction to write_dir_overview() below.
*/
static void write_dir_overview_entry(LIST *startlist)
{
	SCANNODE *current;

	/* Print lines for subdirectories */
	for (current = (SCANNODE *)startlist->first; current; current = (SCANNODE *)current->next)
	{
		if (current->contents)
		{
			/* Print line for current directory */
			fprintf(_fd, "%s\n", current->fullname + strlen(_serverroot));

			/* Process subdirectories as well */
			write_dir_overview_entry(current->contents);
		}
	}
}

/*
** write_dir_overview(startlist)
** 
** Writes an overview of the scanned directories in front of the
** actual filelist.
*/
static void write_dir_overview(LIST *startlist)
{
	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");
	
	fprintf(_fd, "Overview of the directories in this filelist (actual file listing follows):\n\n");

	write_dir_overview_entry(startlist);	

	fprintf(_fd, "\n");
	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");
}

/*
** write_filelist(startlist, name)
**
** Parses the filelist tree and generates entries for the directories and
** files found. "startlist" is the LIST structure of SCANNODEs to be parsed
** and "name" is an associated name. This is a recursive function.
*/
static void write_filelist(LIST *startlist, char *name)
{
	SCANNODE *current;
	BOOL header_written = FALSE;

	/* Process subdirectories first */
	for (current = (SCANNODE *)startlist->first; current; current = (SCANNODE *)current->next)
	{
		if (current->contents)
		{
			/* Write virtual path relative to _serverroot instead of actual filesystem path */
			write_filelist(current->contents, current->fullname + strlen(_serverroot));
		}
	}

	/* Then process the files */
	for (current = (SCANNODE *)startlist->first; current; current = (SCANNODE *)current->next)
	{
		/* We handled subdirectories above, now deal with the current dir's files */
		if (!(IS_DIR(current)))
		{
			/* Write header line for the current directory, if not already done */
			if (!header_written)
			{
				fprintf(_fd, "-----[ %s ]-----\n\n", name);
				header_written = TRUE;
			}
			
			/* Write nickname and filename left-bound */
			sprintf(_buf, "!%s %s", *_nickname, current->name);
			fprintf(_fd, "%-74.74s ::INFO:: ", _buf);

			/* Append file size in human-readable notation */
			if (current->size > 1024*1024)
				sprintf(_buf, "%4.1fM", ((float)current->size / (1024*1024)) );
			else sprintf(_buf, "%4dK", (int)current->size / 1024);
			fprintf(_fd, "%s", _buf);

			/* Update filelist statistics */
			_totalkbytes += (current->size / 1024);
			_totalfiles++;

			/* Append filetype-specific information, if available */
			switch (current->infotype)
			{
				case INFOTYPE_MPEGAUDIO:
					get_mpegaudio_info_string((MPEGAUDIOINFO *)current->info, _buf);
					fprintf(_fd, " %s\n", _buf);
					break;

				case INFOTYPE_WAVEAUDIO:
					get_waveaudio_info_string((WAVEAUDIOINFO *)current->info, _buf);
					fprintf(_fd, " %s\n", _buf);
					break;

				default:
				{
					fprintf(_fd, "\n");
				}
			}
		}
	}

	/* If a header was written for this directory, we also wrote lines for at
	   least one file, so we add another linefeed to optically signalize the end
	   of this directory. */
	if (header_written) fprintf(_fd, "\n");
}

/*
** rc = finish_filelist()
**
** Finishes the filelist with a footer with statistics.
*/
static BOOL finish_filelist(void)
{
	float totalsize;

	errno = 0;

	fprintf(_fd, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
	fprintf(_fd, "\n");

	/* Calculate how long we took */
	get_mins_and_secs(time(NULL) - _starttime, &_mins_passed, &_secs_passed);

	totalsize = ((float)_totalkbytes) / (1024 * 1024);
	if (!totalsize)
	{
		totalsize = ((float)_totalkbytes) / 1024;
		fprintf(_fd, "List totals: %2.2f MB in %d file(s)\n", totalsize, _totalfiles);
	}
	else fprintf(_fd, "List totals: %2.2f GB in %d file(s)\n", totalsize, _totalfiles);
	fprintf(_fd, "List generation took %d minute(s) and %d second(s)\n\n", _mins_passed, _secs_passed);
	fprintf(_fd, "Generated by makelist (MetalServe %s) - http://www.metalserve.org/\n", PACKAGE_VERSION);

	fclose(_fd);
	_fd = NULL;

	if (errno) return FALSE;
	else return TRUE;
}

/*
** rc = compress_filelist()
**
** Processes the configured compress_command and launches it to compress the generated
** filelist. Replaces the placeholders %a with the filename for the compressed archive (without
** suffix) and %u with the filename of the uncompressed list. If compression was successful,
** the uncompressed list is deleted unless the "keep_uncompressed_list" directive was set to
** "true" or "yes".
*/
static BOOL compress_filelist(void)
{
	char archive_name[FILENAME_MAX], *tmp;
	char *p, *q;
	int argslen, pid, status;

	/* Compose archive filename (without suffix) */
	strftime(_buf, sizeof(_buf), "%m-%d-%Y", localtime(&_starttime));
	sprintf(archive_name, "%s-[%s]", *_nickname, _buf);

	/* Generate copy of compress command arguments with replaced placeholders */
	argslen = strlen(_compress_command) + 2 * FILENAME_MAX;
	tmp = malloc(argslen);
	if (!tmp)
	{
		fprintf(stderr, "%s: not enough memory to process compress_command!\n",
		        _prgname);
		return FALSE;
	}

	p = _compress_command;
	q = tmp;
	while ((*p) && ((q - tmp) < argslen))
	{
		if ( (*p == '%') && (*(p+1) == 'a') )
		{
			strncpy(q, archive_name, FILENAME_MAX);
			p += 2;
			q += strlen(archive_name);
		}
		else if ( (*p == '%') && (*(p+1) == 'u') )
		{
			strncpy(q, _filename, FILENAME_MAX);
			p += 2;
			q += strlen(_filename);
		}
		else *q++ = *p++;
	}
	free(_compress_command);
	_compress_command = tmp;

	/* Instead of the interrupt-resistent system() call, we execute the
	   compression program as a child process. */
	pid = fork();

	/* Did something go wrong? */
	if (pid == -1)
	{
		fprintf(stderr, "%s: could not fork() to launch compress_command!\n",
		        _prgname);
		return FALSE;
	}

	/* Are we the child? */
	if (pid == 0)
	{
		char *argv[4];

		/* Redirect in- and output to /dev/null */
		freopen("/dev/null", "r", stdin);
		freopen("/dev/null", "w", stdout);
		freopen("/dev/null", "w", stderr);

		/* Construct argument array for execve() call */
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = _compress_command;
		argv[3] = 0;

		/* execve() only returns if errors occured, and in that case with -1 */
		if (execve("/bin/sh", argv, NULL) == -1)
		{
			fprintf(stderr, "%s: execution of compress_command \"%s\" failed:\n%s!\n",
		        	_prgname, _compress_command, strerror(errno));
			return FALSE;
		}
	}

	/* Wait for child process to exit */
	while (TRUE)
	{
		if (waitpid(pid, &status, 0) == -1)
		{
			if (errno != EINTR)
				return FALSE;
		}
		else
		{
			/* Did the command return an error? */
			if (WEXITSTATUS(status))
			{
				fprintf(stderr, "%s: execution of compress_command \"%s\" failed with return code %d!\n",
			        	_prgname, _compress_command, WEXITSTATUS(status));
				return FALSE;
			}

			/* Delete original text file unless we were told not to do so */
			if (!_keep_uncompressed_list)
			{
				if (remove(_filename))
				{
					fprintf(stderr, "%s: could not remove uncompressed textfile \"%s\":\n%s!\n",
				        	_prgname, _filename, strerror(errno));
					return FALSE;
				}
			}

			return TRUE;
		}
	}
}

/*
** Main routine
*/
int main(int argc, char **argv)
{
	BOOL rc;

	/* Determine program name */
	_prgname = (char *)basename(strdup(argv[0]));

	/* Parse command line arguments */
	if (!parse_cmdline_args(argc, argv))
		return 1;

	/* Parse configuration file */
	if (!parse_cfgfile())
		return 1;

	if (!_quiet)
	{
		/* Print program banner */
		printf(_prgbanner, _prgname, PACKAGE_NAME, PACKAGE_VERSION);

		/* Determine whether a terminal is attached */
		if (isatty(STDOUT_FILENO))
		{
			/* Determine terminal width */
			_terminal_width = determine_terminal_width();
			if (!_terminal_width)
				return 1;

			_tty_attached = TRUE;
		}
	}

	/* Allocate memory for the filelist */
	_filelist = allocate_list();
	if (!_filelist)
	{
		fprintf(stderr, "%s: not enough memory for filelist!\n",
		        _prgname);
		return FALSE;
	}

	/* Remember starttime for statistical purposes */
	_starttime = time(NULL);

	/* Process the directory tree and add files to the list */
	if (!_quiet)
		printf("Scanning \"%s\" ...\n", _serverroot);

	rc = process_directory(_filelist, _serverroot);

	/* Start the on-disk filelist */
	if (rc)
	{
		if (!_quiet)
		{
			printf("\r%-*s\n", _terminal_width, "Writing filelist ...");
			fflush(stdout);
		}

		rc = start_filelist();
	}

	/* Write directory overview, if enabled */
	if ( (rc) && (_dir_overview) )
		write_dir_overview(_filelist);

	/* Write files from LIST structure to on-disk filelist */
	if (rc)
		write_filelist(_filelist, "/");

	/* Finish on-disk filelist */
	if (rc)
		rc = finish_filelist();

	/* Compress filelist, if enabled */
	if ( (rc) && (_compress_command) )
	{
		if (!_quiet)
		{
			printf("Compressing filelist ...\n");
			fflush(stdout);
		}

		if (!compress_filelist())
			return 1;
	}

	/* If errors occured, attempt to delete the incomplete textfile */
	if (!rc)
	{
		if (_fd)
			fclose(_fd);
		unlink(_filename);	
	
		return 1;
	}

	/* Print statistics */
	if (!_quiet)
		printf("\nTotal time elapsed: %02d:%02d\n", _mins_passed, _secs_passed);

	return 0;
}
